#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
A program to download my Lepidoptera data and generate a webpage!

Created on Fri Aug 16 16:23:40 2019
Version : 0.5 beta - 25 Aug 2019
@author: Bhalchandra Pujari 
@email : bspujari@cms.unipune.ac.in



"""




username="bspujari"
location="Pune"


#--- MAIN CODE BEGIN NOW

from halo import Halo #spinnner for loop
import requests
import json
import math,random
import time,datetime
import os.path,sys
import collections
from collections import defaultdict
#import matplotlib.pyplot as plt
import pandas as pd
#from bokeh.io import output_file, show
#from bokeh.palettes import Cividis256
from bokeh.plotting import figure
from bokeh.transform import cumsum
from bokeh.embed import autoload_static
from bokeh.layouts import row
from bokeh.resources import CDN
#from bokeh.models import Legend








class iNat_leps():
    
    def main(self):
        
        print ("Initializing ..")
        self.name_rank_dict=defaultdict(list)
        self.taxa_id_dump()

        print ('Downloading the data ... hang on..')
        self.iNat_data=[] # empty list to save data
        self.parse_iNat() # fetch the data from iNat and save in above list
        
        print ("Extracting the data..")
        self.extract_image_data() # extract the relevent data and 
        
    
        print ("Plotting ..")
        self.plots()

        
        print ("Creating the webpage..")
        self.write_html("welcome.html")
        
        print ("Everything done. Please check 'welcome.html'. Good bye.")

            
#-----MAIN ENDS HERE -----------------------------------


        
    def taxa_id_dump(self): 
        self.taxafile="taxa_id.csv" # a dump file to store taxa-id and names"
        self.TaxaId_Name=pd.DataFrame(columns=("TaxaID", "Name", "RankName", "RankLevel")) # a grand dict storing all IDs and corresponding names. 
        if os.path.isfile(self.taxafile): 
            print ("Loaiding")
            self.TaxaId_Name =pd.read_csv(self.taxafile, \
                    names=("TaxaID", "Name", "RankName", "RankLevel")\
                    )
        else:
            print("Not exist")
            
        self.TaxaId_Name.astype({'TaxaID': 'int64'})
#        if os.path.isfile(self.taxafile): # read tha database of id-tax available on disk. If nothing is available it will create a database.
#            with open(self.taxafile) as f:
#                for line in f:
#                    (key, val) = line.split()
#                    self.TaxaId_Name[int(key)] = val
                    
    def plots(self): # make plots using bokeh
         plot_items=['family','tribe','genus','species']
         plots=[]
         for item in plot_items :
             pie_data=collections.Counter(self.name_rank_dict[item])
             data=pd.Series(pie_data).reset_index(name='value').rename(columns={'index':item})
             data['angle'] = data['value']/data['value'].sum() * 2*math.pi
             data['color'] = ["#"+''.join([random.choice('0123456789ABCDEF') for j in range(6)]) for i in range(len(data))]
             #p = figure(plot_width=275, plot_height=250, title=item.upper()+" ("+str(len(data))+")", toolbar_location=None,
             p = figure(width=275, height=250, title=item.upper()+" ("+str(len(data))+")", toolbar_location=None,
                        tools="hover", tooltips="@"+item+": @value", sizing_mode='scale_width')
             p.wedge(x=0, y=1, radius=0.8,
                     start_angle=cumsum('angle', include_zero=True), end_angle=cumsum('angle'),
                     line_color="white", fill_color='color', 
                     source=data)
             p.axis.visible = False
             p.min_border=0
             plots.append(p)
             
         js_dest="dist/js/bokeh.js"
         js, tag = autoload_static(row(plots), CDN, js_dest)
         f = open(js_dest,"w")
         f.write(js)
         f.close()
         self.htmlplots="<center>"+tag+"</center><br><hr><br>\n\n"

        
    def write_html(self,fwrite):
        f= open(fwrite,"w")
        header=open("header.tx","r")
        footer=open("footer.tx","r")
        f.write(header.read())
        f.write(self.htmlplots)
        f.write("<div class=\"mainbox\">\n<h3> Identified taxa (click to expand) </h3>")
        plot_items=['family','tribe','genus','species']
        for item in plot_items :
            f.write("<div class =\"childdiv\">")
            f.write("<button class=\"collapsible\">"+item.upper()+"</button>")
            f.write("<div class=\"content\">")
            f.write("<p>")
            string=" <br>".join(sorted(set(self.name_rank_dict[item])))
            f.write(string)
            f.write("</p>")        
            f.write("</div>")
            f.write("</div>")
            
        f.write("</div> <hr><br>")
        f.write("<h3> Gallery </h3>")
        f.write(self.html_image_gallery)
        now = datetime.datetime.now().strftime("%d/%B/%Y %H:%M")
        f.write(footer.read().replace('UNKNOWN',now))
        
        f.close()
        footer.close()
        
        
        
    def extract_image_data(self):
        self.html_image_gallery="" # text to be appended in html file
        uniq_sci_name=[]
        spinner = Halo( text_color='blue',spinner='line')
        f=open(self.taxafile,'a')  # keep open the taxa_name file to append the unknonwn taxa_name pair
        #rf=open(self.rankdump,'a') # file fo dump rank and rank levels
        
        for entry in self.iNat_data:
            spinner.start(text="Updating the database (takes looong at the first instance)")
            for i in range(len(entry['results'])):
                if entry['results'][i]['place_guess'] and  location in entry['results'][i]['place_guess']: # Extract  data only from within SPPU
                    ancestry=[int(z) for z in entry['results'][i]['taxon']["min_species_ancestry"].split(",")]
                    if 47157 in ancestry : # GET only Lepidoptera (taxa id=47157)
                        obs_id=entry['results'][i]['id'] 
                        obs_url="https://www.inaturalist.org/observations/"+str(obs_id)
                        sci_name=entry['results'][i]['taxon']['name'] 
                        rank=entry['results'][i]['taxon']['rank'] # it is a string corresponding to rank_level ('genus'/'family' etc)
                        rank_level=entry['results'][i]['taxon']['rank_level']
                        # ^^^^^^^ what is the ID rank? Is is species or genus or family?
                        taxa_id=entry['results'][i]['taxon']['min_species_taxon_id'] # taxa ID. 
                        
                        thumb_url=entry['results'][i]['observation_photos'][0]['photo']['url'].split('?')[0]
                        original_url=thumb_url.replace("square","original")
                        
                        if rank_level == 10:
                            if sci_name not in uniq_sci_name: # gallery for uniq 'species'only.. but accept multiple for higher ranks
                                uniq_sci_name.append(sci_name)
                                #print (sci_name, rank,rank_level)
                                self.html_image_gallery = self.html_image_gallery +"<a href=\""+original_url+\
                                   "\" data-lumenbox=\"my-gallery\"  data-title=\""+sci_name+" ("+rank+") <a href=\'"+obs_url+"\' target=\'new\'> Click for details </a> \"> <img src=\""+thumb_url+\
                                    "\" alt=\""+sci_name+"\" /></a>\n"
                        else:
                            self.html_image_gallery = self.html_image_gallery +"<a href=\""+original_url+\
                               "\" data-lumenbox=\"my-gallery\"  data-title=\""+sci_name+" ("+rank+") <a href=\'"+obs_url+"\' target=\'new\'> Click for details </a>\" > <img src=\""+thumb_url+\
                                "\" alt=\""+sci_name+"\" /></a>\n"
                                
                        #create a dict of ranks.. not uniq .. species/genus/family etc.. we'll plot this for abundance 
                        #print (self.TaxaId_Name)
                        #print(self.TaxaId_Name.describe())
                        #if not self.TaxaId_Name['TaxaID'].str.contains(str(taxa_id)).any():
                        #    newrow={"TaxaID":str(taxa_id), "Name":sci_name, "RankName":rank, "RankLevel":rank_level}
                        #    self.TaxaId_Name=self.TaxaId_Name.append(newrow, ignore_index=True)
                        # do a thourough search of taxa for name of all the ancesters (like tibe, family etc)
                        for ancester in ancestry :
                            #print(ancester)
                            
                            #if not self.TaxaId_Name['TaxaID'].str.contains(str(ancester)).any():
                            #if not ancester in  self.TaxaId_Name['TaxaID']:
                            #print ("DEBUG", self.TaxaId_Name[self.TaxaId_Name["TaxaID"]==ancester])
                            if len(self.TaxaId_Name[self.TaxaId_Name["TaxaID"]==ancester]) == 0: # means no entry for ancester is found in database
                                api_string="http://api.inaturalist.org/v1/taxa/"+str(ancester)
                                r=requests.get(api_string) # download .. make sure proxies are fine

                                try:
                                    r=json.loads(r.text)
                                except :
                                    print ("Error:", ancester, r)
                                if not r :
                                    print ("I tried ", api_string , " and failed")
                                taxaname=r['results'][0]['name']                                    
                                rank=r['results'][0]['rank']
                                rank_level=r['results'][0]['rank_level']
                                newrow={"TaxaID":ancester, "Name":taxaname, "RankName":rank, "RankLevel":rank_level}
                                #self.TaxaId_Name=self.TaxaId_Name.append(newrow, ignore_index=True)
                                self.TaxaId_Name = pd.concat([self.TaxaId_Name, pd.DataFrame([newrow])], ignore_index=True)
                                #update databases
                                #self.rank_ranklevel[taxalevel] = taxarank
                                #rf.write(str(taxalevel) +" "+ str(taxarank))
                                
                                #self.TaxaId_Name[int(ancester)]= taxaname
                                f.write(str(ancester)+","+taxaname+","+str(rank) +"," + str(rank_level)+"\n")
                            else:
                                rank=self.TaxaId_Name[self.TaxaId_Name['TaxaID']==ancester]['RankName'].values[0]
                                taxaname=self.TaxaId_Name[self.TaxaId_Name['TaxaID']==ancester]['Name'].values[0]
                                
                            self.name_rank_dict[rank].append(taxaname)       
                            
                        #if sci_name not in self.name_rank_dict[rank]:
                        
                        
        f.close()
        #rf.close()
        spinner.succeed("Extraction done")
        #print (self.TaxaId_Name)
            
            

    def parse_iNat(self): 
      
      page_count=1 # Required by the format of API 
           # total observations = pages * observations per page
           #                  (set to 100 at this stage)
      total_pages='NA'
      
      
      
      spinner = Halo( text_color='green',spinner='clock')
      t0=time.time()
      while True:
      #while page_count <2:
          
          if page_count == 1 : 
              spin_text="Page "+str(page_count)
          else :
              spin_text="Page "+str(page_count)+" of "+ total_pages+\
              "  Time Remaining : " + str(math.ceil((t1-t0)*(int(total_pages)-page_count+1))) + " sec"
          spinner.start(text=spin_text )
          

              # Build the string for API download
          api_string="http://api.inaturalist.org/v1/observations?"+\
          "&identified=true"+\
          "&photos=true"+\
          "&page="+str(page_count)+\
          "&per_page=100"+\
          "&order=desc"+\
          "&user_id="+username+\
          "&order_by=created_at"
          # looks like 100 is the limit for per page observations.
          
          try:
              r=requests.get(api_string) # download .. make sure proxies are fine
          except :
              print ("\nERROR: Active internet connection is required to download the data. Please check you connection and try again.\nI have to stop here. Bye.")
              sys.exit()
              
          j=json.loads(r.text)
          
          
          if page_count == 1 : # for the first iteration, determine how many pages of downloadng is there..
              total_pages=str(math.ceil(j['total_results']/100) +1) # total results/results per page
              t1=time.time()
              
          if j['results'] : # error will occur when page numbers are exhausted.. till that time keep increasing page numbers.
              page_count+=1
              self.iNat_data.append(json.loads(r.text))
          else:
              break
          
          
          spinner.stop()
      spinner.succeed("Download complete") 
if __name__ == '__main__':
   u=iNat_leps()
   u.main()
